using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Net.Http;
using System.Threading;
using System.Threading.Tasks;
using Props.Shop.Framework;

namespace Props.Shop.Adafruit.Api
{
    public class ProductListingManager : IProductListingManager
    {
        public event EventHandler DataUpdateEvent;
        private bool disposedValue;
        private volatile Dictionary<string, IList<ProductListing>> activeListings;
        public IDictionary<string, IList<ProductListing>> ActiveListings => activeListings;
        private ProductListingsParser parser = new ProductListingsParser();
        private HttpClient httpClient;
        private Timer updateTimer;

        public ProductListingManager(HttpClient httpClient)
        {
            this.httpClient = httpClient;
        }

        public async Task DownloadListings() {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            HttpResponseMessage responseMessage = await httpClient.GetAsync("products");
            parser.BuildProductListings(responseMessage.Content.ReadAsStream());
            Dictionary<string, IList<ProductListing>> listings = new Dictionary<string, IList<ProductListing>>();
            foreach (ProductListing product in parser.ProductListings)
            {
                IList<ProductListing> sameProducts = listings.GetValueOrDefault(product.Name);
                if (sameProducts == null) {
                    sameProducts = new List<ProductListing>();
                    listings.Add(product.Name, sameProducts);
                }

                sameProducts.Add(product);
            }
            activeListings = listings;
            DataUpdateEvent?.Invoke(this, null);
        }

        public void StartUpdateTimer(int delay = 1000 * 60 * 5, int period = 1000 * 60 * 5) {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (updateTimer != null) throw new InvalidOperationException("Update timer already started.");
            updateTimer = new Timer(async (state) => await DownloadListings(), null, delay, period);
        }

        public void StopUpdateTimer() {
            if (disposedValue) throw new ObjectDisposedException("ProductListingManager");
            if (updateTimer != null) throw new InvalidOperationException("Update timer not started.");
            updateTimer.Dispose();
            updateTimer = null;
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    updateTimer?.Dispose();
                    updateTimer = null;
                }
                disposedValue = true;
            }
        }

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
    }
}